# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Loop definitions, inspired by DeepMind Acme."""

from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Sequence

from absl import logging

from oatomobile.core.agent import Agent
from oatomobile.core.rl import Env
from oatomobile.core.rl import Metric
from oatomobile.core.typing import Scalar


class EnvironmentLoop:
  """A simple RL-like environment loop.

  This takes `Env` and `Agent` instances and coordinates their
  interaction. This can be used as:

    loop = EnvironmentLoop(environment, agent_fn)
    loop.run()
  """

  def __init__(
      self,
      agent_fn: Callable[..., Agent],
      environment: Env,
      metrics: Optional[Sequence[Metric]] = None,
      render_mode: str = "none",
  ) -> None:
    """Constructs an environment loop.

    Args:
      agent_fn: The agent's construction function that receives each task.
      environment: The environment to evaluate on.
      metrics: Set of metrics to record during episode.
      render_mode: The render mode, one of {"none", "human", "rgb_array"}.
    """
    assert render_mode in ("none", "human", "rgb_array")

    # Internalizes agent and environment.
    self._agent_fn = agent_fn
    self._environment = environment
    self._metrics = metrics
    self._render_mode = render_mode

  def run(self) -> Optional[Mapping[str, Scalar]]:
    """Perform the run loop.

    Returns:
      If `metrics` are provided, it returns their final values in a dictionary
      format.
    """

    try:
      # Initializes environment.
      done = False
      observation = self._environment.reset()
      if self._render_mode is not "none":
        self._environment.render(mode=self._render_mode)
      # Initializes agent.
      agent = self._agent_fn(environment=self._environment)

      # Episode loop.
      while not done:
        # Get vehicle control.
        action = agent.act(observation)

        # Progresses the simulation.
        new_observation, reward, done, _ = self._environment.step(action)
        if self._render_mode is not "none":
          self._environment.render(mode=self._render_mode)

        # Updates the agent belief.
        agent.update(observation, action, new_observation)

        # Update metrics.
        if self._metrics is not None:
          for metric in self._metrics:
            metric.update(observation, action, reward, new_observation)

        # Book-keeping.
        observation = new_observation

    except Exception as msg:
      logging.error(msg)

    finally:
      # Garbage collector.
      try:
        environment.close()
      except NameError:
        pass

      # Returns the recorded metrics.
      if self._metrics is not None:
        return {metric.uuid: metric.value for metric in self._metrics}
      else:
        return None



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Logging helper functions for `TensorFlow` baselines."""

import os
from typing import Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import sonnet as snt
import tensorflow as tf

from oatomobile.baselines.tf.typing import ArrayLike

COLORS = [
    "#0071bc",
    "#d85218",
    "#ecb01f",
    "#7d2e8d",
    "#76ab2f",
    "#4cbded",
    "#a1132e",
]


class TensorBoardWriter:
  """A simple `TensorFlow`-friendly `TensorBoard` wrapper."""

  def __init__(
      self,
      log_dir: str,
  ) -> None:
    """Constructs a simgple `TensorBoard` wrapper."""
    # Makes sure output directories exist.
    log_dir_train = os.path.join(log_dir, "train")
    log_dir_val = os.path.join(log_dir, "val")
    os.makedirs(log_dir_train, exist_ok=True)
    os.makedirs(log_dir_val, exist_ok=True)

    # Initialises the `TensorBoard` writters.
    self._summary_writter_train = tf.summary.create_file_writer(log_dir_train)
    self._summary_writter_val = tf.summary.create_file_writer(log_dir_val)

  def log(
      self,
      split: str,
      loss: float,
      overhead_features: ArrayLike,
      predictions: ArrayLike,
      ground_truth: ArrayLike,
      global_step: int,
  ) -> None:
    """Logs the scalar loss and visualizes predictions for qualitative
    inspection.

    Args:
      split: One of {"train", "val"}, raises error otherwise.
      loss: The optimization objective value.
      overhead_features: The overhead visual input, e.g., LIDAR, with shape `[B, H, W, 3]`.
      predictions: Samples from the model, with shape `[B, (K,) T, 2]`.
      ground_truth: The ground truth plan, with shape `[B, T, 2]`.
      global_step: The x-axis value.
    """

    if split == "train":
      summary_writter = self._summary_writter_train
    elif split == "val":
      summary_writter = self._summary_writter_val
    else:
      raise ValueError("Unrecognised split={} was passed".format(split))

    with summary_writter.as_default():
      tf.summary.scalar("loss", data=loss, step=global_step)

      # Visualizes the predictions.
      raw = list()
      for t, (
          o_t,
          p_t,
          g_t,
      ) in enumerate(zip(
          overhead_features,
          predictions,
          ground_truth,
      )):
        fig, ax = plt.subplots(figsize=(3.0, 3.0))
        bev_meters = 25.0
        # Overhead features.
        ax.imshow(
            o_t.squeeze(),
            extent=(-bev_meters, bev_meters, bev_meters, -bev_meters),
            cmap="gray",
        )
        # Ground truth.
        ax.plot(
            g_t[..., 1],
            -g_t[..., 0],
            marker="o",
            markersize=4,
            color=COLORS[1],
            alpha=1.0,
            label="ground truth",
        )
        # Model prediction.
        ax.plot(
            p_t[..., 1],
            -p_t[..., 0],
            marker="o",
            markersize=4,
            color=COLORS[0],
            alpha=1.0,
            label="predictions",
        )
        ax.legend()
        ax.set(frame_on=False)
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        # Convert `matplotlib` canvas to `NumPy` RGB.
        fig.canvas.draw()
        w, h = fig.canvas.get_width_height()
        buf = np.fromstring(fig.canvas.tostring_argb(), dtype=np.uint8)
        buf.shape = (w, h, 4)
        buf = np.roll(buf, 3, axis=2)
        raw.append(buf)
        plt.close(fig)
      raw = np.reshape(np.asarray(raw), (-1, w, h, 4))
      tf.summary.image("examples", raw, max_outputs=16, step=global_step)



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""CARLA utility functions and wrappers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import atexit
import collections
import os
import random
import signal
import subprocess
import sys
import time
from typing import Any
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import carla
import numpy as np
import transforms3d.euler
from absl import logging


def setup(
    town: str,
    fps: int = 20,
    server_timestop: float = 20.0,
    client_timeout: float = 20.0,
    num_max_restarts: int = 5,
) -> Tuple[carla.Client, carla.World, int, subprocess.Popen]:  # pylint: disable=no-member
  """Returns the `CARLA` `server`, `client` and `world`.

  Args:
    town: The `CARLA` town identifier.
    fps: The frequency (in Hz) of the simulation.
    server_timestop: The time interval between spawing the server
      and resuming program.
    client_timeout: The time interval before stopping
      the search for the carla server.
    num_max_restarts: Number of attempts to connect to the server.

  Returns:
    client: The `CARLA` client.
    world: The `CARLA` world.
    frame: The synchronous simulation time step ID.
    server: The `CARLA` server.
  """
  assert town in ("Town01", "Town02", "Town03", "Town04", "Town05")

  # The attempts counter.
  attempts = 0

  while attempts < num_max_restarts:
    logging.debug("{} out of {} attempts to setup the CARLA simulator".format(
        attempts + 1, num_max_restarts))

    # Random assignment of port.
    port = np.random.randint(2000, 3000)

    # Start CARLA server.
    env = os.environ.copy()
    env["SDL_VIDEODRIVER"] = "offscreen"
    env["SDL_HINT_CUDA_DEVICE"] = "0"
    logging.debug("Inits a CARLA server at port={}".format(port))
    server = subprocess.Popen(
        [
            os.path.join(os.environ.get("CARLA_ROOT"), "CarlaUE4.sh"),
            "-carla-rpc-port={}".format(port),
            "-quality-level=Epic",
        ],
        stdout=None,
        stderr=subprocess.STDOUT,
        preexec_fn=os.setsid,
        env=env,
    )
    atexit.register(os.killpg, server.pid, signal.SIGKILL)
    time.sleep(server_timestop)

    # Connect client.
    logging.debug("Connects a CARLA client at port={}".format(port))
    try:
      client = carla.Client("localhost", port)  # pylint: disable=no-member
      client.set_timeout(client_timeout)
      client.load_world(map_name=town)
      world = client.get_world()
      world.set_weather(carla.WeatherParameters.ClearNoon)  # pylint: disable=no-member
      frame = world.apply_settings(
          carla.WorldSettings(  # pylint: disable=no-member
              no_rendering_mode=False,
              synchronous_mode=True,
              fixed_delta_seconds=1.0 / fps,
          ))
      logging.debug("Server version: {}".format(client.get_server_version()))
      logging.debug("Client version: {}".format(client.get_client_version()))
      return client, world, frame, server
    except RuntimeError as msg:
      logging.debug(msg)
      attempts += 1
      logging.debug("Stopping CARLA server at port={}".format(port))
      os.killpg(server.pid, signal.SIGKILL)
      atexit.unregister(lambda: os.killpg(server.pid, signal.SIGKILL))

  logging.debug(
      "Failed to connect to CARLA after {} attempts".format(num_max_restarts))
  sys.exit()


def carla_rgb_image_to_ndarray(image: carla.Image) -> np.ndarray:  # pylint: disable=no-member
  """Returns a `NumPy` array from a `CARLA` RGB image.

  Args:
    image: The `CARLA` RGB image.

  Returns:
    A `NumPy` array representation of the image.
  """
  image.convert(carla.ColorConverter.Raw)  # pylint: disable=no-member
  array = np.frombuffer(image.raw_data, dtype=np.dtype("uint8"))
  array = array.astype(np.float32) / 255
  array = np.reshape(array, (image.height, image.width, 4))
  array = array[:, :, :3]
  array = array[:, :, ::-1]
  return array


def carla_cityscapes_image_to_ndarray(image: carla.Image) -> np.ndarray:  # pylint: disable=no-member
  """Returns a `NumPy` array from a `CARLA` semantic segmentation image.

  Args:
    image: The `CARLA` semantic segmented image.

  Returns:
    A `NumPy` array representation of the image.
  """
  image.convert(carla.ColorConverter.CityScapesPalette)  # pylint: disable=no-member
  array = np.frombuffer(image.raw_data, dtype=np.dtype("uint8"))
  array = array.astype(np.float32) / 255
  array = np.reshape(array, (image.height, image.width, 4))
  array = array[:, :, :3]
  array = array[:, :, ::-1]
  return array


def carla_lidar_measurement_to_ndarray(
    lidar_measurement: carla.LidarMeasurement,  # pylint: disable=no-member
    pixels_per_meter: int = 2,
    hist_max_per_pixel: int = 5,
    meters_max: int = 50,
) -> np.ndarray:
  """Returns a `NumPy` array from a `CARLA` LIDAR point cloud.

  Args:
    lidar_measurement: The `CARLA` LIDAR point cloud.

  Returns:
    A `NumPy` array representation of the point cloud.
  """

  def splat_points(
      point_cloud,
      pixels_per_meter: int,
      hist_max_per_pixel: int,
      meters_max: int,
  ):
    """Converts point cloud to 2D histograms."""
    # Allocate 2D histogram bins.
    ymeters_max = meters_max
    xbins = np.linspace(
        -meters_max,
        meters_max + 1,
        meters_max * 2 * pixels_per_meter + 1,
    )
    ybins = np.linspace(
        -meters_max,
        ymeters_max + 1,
        ymeters_max * 2 * pixels_per_meter + 1,
    )
    # Compute histogram of x and y coordinates of points.
    hist = np.histogramdd(point_cloud[..., :2], bins=(xbins, ybins))[0]
    # Clip histogram
    hist[hist > hist_max_per_pixel] = hist_max_per_pixel
    # Normalize histogram by the maximum number of points in a bin we care about.
    overhead_splat = hist / hist_max_per_pixel
    # Return splat in X x Y orientation, with X parallel to car axis, Y perp, both parallel to ground.
    return overhead_splat

  # Serialise and parse to `NumPy` tensor.
  points = np.frombuffer(lidar_measurement.raw_data, dtype=np.dtype("f4"))
  points = np.reshape(points, (int(points.shape[0] / 3), 3))

  # Split observations in the Z dimension (height).
  below = points[points[..., 2] <= -2.5]
  above = points[points[..., 2] >= -2.5]
  # Convert point clouds to 2D histograms.
  features = list()
  features.append(
      splat_points(
          below,
          pixels_per_meter,
          hist_max_per_pixel,
          meters_max,
      ))
  features.append(
      splat_points(
          above,
          pixels_per_meter,
          hist_max_per_pixel,
          meters_max,
      ))
  features = np.stack(features, axis=-1)

  return features.astype(np.float32)


def spawn_hero(
    world: carla.World,  # pylint: disable=no-member
    spawn_point: carla.Transform,  # pylint: disable=no-member
    vehicle_id: Optional[str] = None,
) -> carla.Vehicle:  # pylint: disable=no-member
  """Spawns `hero` in `spawn_point`.

  Args:
    world: The world object associated with the simulation.
    spawn_point: The point to spawn the hero actor.
    vehicle_id: An (optional) valid id for the vehicle object.

  Returns:
    The actor (vehicle) object.
  """
  # Blueprints library.
  bl = world.get_blueprint_library()
  if vehicle_id is not None:
    # Get the specific vehicle from the library.
    hero_bp = bl.find(vehicle_id)
  else:
    # Randomly choose a vehicle from the list.
    hero_bp = random.choice(bl.filter("vehicle.*"))
  # Rename the actor to `hero`.
  hero_bp.set_attribute("role_name", "hero")
  logging.debug("Spawns hero actor at {}".format(
      carla_xyz_to_ndarray(spawn_point.location)))
  hero = world.spawn_actor(hero_bp, spawn_point)
  return hero


def spawn_vehicles(
    world: carla.World,  # pylint: disable=no-member
    num_vehicles: int,
) -> Sequence[carla.Vehicle]:  # pylint: disable=no-member
  """Spawns `vehicles` randomly in spawn points.

  Args:
    world: The world object associated with the simulation.
    num_vehicles: The number of vehicles to spawn.

  Returns:
    The list of vehicles actors.
  """
  # Blueprints library.
  bl = world.get_blueprint_library()
  # List of spawn points.
  spawn_points = world.get_map().get_spawn_points()
  # Output container
  actors = list()
  for _ in range(num_vehicles):
    # Fetch random blueprint.
    vehicle_bp = random.choice(bl.filter("vehicle.*"))
    # Attempt to spawn vehicle in random location.
    actor = world.try_spawn_actor(vehicle_bp, random.choice(spawn_points))
    if actor is not None:
      # Enable autopilot.
      actor.set_autopilot(True)
      # Append actor to the list.
      actors.append(actor)
  logging.debug("Spawned {} other vehicles".format(len(actors)))
  return actors


def spawn_pedestrians(
    world: carla.World,  # pylint: disable=no-member
    num_pedestrians: int,
    speeds: Sequence[float] = (1.0, 1.5, 2.0),
) -> Sequence[carla.Vehicle]:  # pylint: disable=no-member
  """Spawns `pedestrians` in random locations.

  Args:
    world: The world object associated with the simulation.
    num_pedestrians: The number of pedestrians to spawn.
    speeds: The valid set of speeds for the pedestrians.

  Returns:
    The list of pedestrians actors.
  """
  # Blueprints library.
  bl = world.get_blueprint_library()
  # Output container
  actors = list()
  for n in range(num_pedestrians):
    # Fetch random blueprint.
    pedestrian_bp = random.choice(bl.filter("walker.pedestrian.*"))
    # Make pedestrian invicible.
    pedestrian_bp.set_attribute("is_invincible", "true")
    while len(actors) != n:
      # Get random location.
      spawn_point = carla.Transform()  # pylint: disable=no-member
      spawn_point.location = world.get_random_location_from_navigation()
      if spawn_point.location is None:
        continue
      # Attempt to spawn vehicle in random location.
      actor = world.try_spawn_actor(pedestrian_bp, spawn_point)
      if actor is not None:
        actors.append(actor)
  logging.debug("Spawned {} pedestrians".format(len(actors)))
  return actors


def spawn_camera(
    hero: carla.ActorBlueprint,  # pylint: disable=no-member
    config: Mapping[str, Any],
    camera_type: str,
) -> carla.ServerSideSensor:  # pylint: disable=no-member
  """Spawns a camera on `hero`.

  Args:
    hero: The agent to attach the camera on.
    config: The attribute-value pairs for the configuration
      of the sensor.
    camera_type: Camera type, one of ("rgb", "semantic_segmentation").

  Returns:
    The spawned  camera sensor.
  """
  assert camera_type in ("rgb", "semantic_segmentation")

  # Get hero's world.
  world = hero.get_world()
  # Blueprints library.
  bl = world.get_blueprint_library()
  # Configure blueprint.
  camera_bp = bl.find("sensor.camera.{}".format(camera_type))
  for attribute, value in config["attributes"].items():
    camera_bp.set_attribute(attribute, value)
  logging.debug("Spawns a {} camera".format(camera_type))
  return world.spawn_actor(
      camera_bp,
      carla.Transform(  # pylint: disable=no-member
          carla.Location(**config["actor"]["location"]),  # pylint: disable=no-member
          carla.Rotation(**config["actor"]["rotation"]),  # pylint: disable=no-member
      ),
      attach_to=hero,
  )


def spawn_lidar(
    hero: carla.ActorBlueprint,  # pylint: disable=no-member
    config: Mapping[str, Any],
) -> carla.ServerSideSensor:  # pylint: disable=no-member
  """Spawns LIDAR sensor on `hero`.

  Args:
    hero: The agent to attach the LIDAR sensor on.
    config: The attribute-value pairs for the configuration
      of the sensor.

  Returns:
    The spawned LIDAR sensor.
  """
  # Get hero's world.
  world = hero.get_world()
  # Blueprints library.
  bl = world.get_blueprint_library()
  # Configure blueprint.
  lidar_bp = bl.find("sensor.lidar.ray_cast")
  for attribute, value in config["attributes"].items():
    lidar_bp.set_attribute(attribute, value)
  logging.debug("Spawns a LIDAR sensor")
  return world.spawn_actor(
      lidar_bp,
      carla.Transform(  # pylint: disable=no-member
          carla.Location(**config["actor"]["location"]),  # pylint: disable=no-member
          carla.Rotation(),  # pylint: disable=no-member
      ),
      attach_to=hero,
  )


def spawn_collision(
    hero: carla.ActorBlueprint,  # pylint: disable=no-member
) -> carla.ServerSideSensor:  # pylint: disable=no-member
  """Spawns collision sensor on `hero`.

  Args:
    hero: The agent to attach the collision sensor on.

  Returns:
    The spawned collision sensor.
  """
  # Get hero's world.
  world = hero.get_world()
  # Blueprints library.
  bl = world.get_blueprint_library()
  # Configure blueprint.
  collision_bp = bl.find("sensor.other.collision")
  logging.debug("Spawns a collision sensor")
  return world.spawn_actor(
      collision_bp,
      carla.Transform(),  # pylint: disable=no-member
      attach_to=hero,
  )


def spawn_lane_invasion(
    hero: carla.ActorBlueprint,  # pylint: disable=no-member
) -> carla.ServerSideSensor:  # pylint: disable=no-member
  """Spawns lane invasion sensor on `hero`.

  Args:
    hero: The agent to attach the collision sensor on.

  Returns:
    The spawned lane invasion sensor.
  """
  # Get hero's world.
  world = hero.get_world()
  # Blueprints library.
  bl = world.get_blueprint_library()
  # Configure blueprint.
  collision_bp = bl.find("sensor.other.lane_invasion")
  logging.debug("Spawns a lane invasion sensor")
  return world.spawn_actor(
      collision_bp,
      carla.Transform(),  # pylint: disable=no-member
      attach_to=hero,
  )


def get_spawn_point(
    world: carla.World,  # pylint: disable=no-member
    spawn_point: Optional[Union[int, carla.Transform]]  # pylint: disable=no-member
) -> carla.Location:  # pylint: disable=no-member
  """Parses and returns a CARLA spawn points."""
  if isinstance(spawn_point, carla.Transform):  # pylint: disable=no-member
    _spawn_point = spawn_point
  elif isinstance(spawn_point, int):
    _spawn_point = world.get_map().get_spawn_points()[spawn_point]
  else:
    _spawn_point = random.choice(world.get_map().get_spawn_points())
  return _spawn_point


def get_actors(
    world: carla.World,  # pylint: disable=no-member
    spawn_point: Optional[Union[int, carla.Location]],  # pylint: disable=no-member
    num_vehicles: int,
    num_pedestrians: int,
) -> Tuple[carla.Vehicle, Sequence[Optional[carla.Vehicle]],  # pylint: disable=no-member
           Sequence[Optional[carla.Walker]]]:  # pylint: disable=no-member
  """Spawns and returns the `hero`, the `vehicles` and the `pedestrians`.

  Args:
    world: The world object associated with the simulation.
    spawn_point: The hero vehicle spawn point. If an int is
      provided then the index of the spawn point is used.
      If None, then randomly selects a spawn point every time
      from the available spawn points of each map.
    num_vehicles: The number of vehicles to spawn.
    num_pedestrians: The number of pedestrians to spawn.

  Returns:
    hero: The spawned ego vehicle agent object.
    vehicles: The spawned vehicles agent objcets.
    pedestrians: The spawned walker agent objects.
  """
  # HERO agent.
  _spawn_point = get_spawn_point(world, spawn_point)
  hero = spawn_hero(
      world=world,
      spawn_point=_spawn_point,
      vehicle_id="vehicle.ford.mustang",
  )
  # Other vehicles.
  vehicles = spawn_vehicles(
      world=world,
      num_vehicles=num_vehicles,
  )
  # Other pedestrians.
  pedestrians = spawn_pedestrians(
      world=world,
      num_pedestrians=num_pedestrians,
  )
  return hero, vehicles, pedestrians


def vehicle_to_carla_measurements(
    vehicle: carla.Vehicle,  # pylint: disable=no-member
) -> Mapping[str, Any]:
  """Wraps all the `get_` calls from the `CARLA` interface."""
  control = vehicle.get_control()
  _transform = vehicle.get_transform()
  location = _transform.location
  rotation = _transform.rotation
  velocity = vehicle.get_velocity()
  acceleration = vehicle.get_acceleration()
  orientation = _transform.get_forward_vector()
  angular_velocity = vehicle.get_angular_velocity()
  speed_limit = vehicle.get_speed_limit()
  is_at_traffic_light = vehicle.is_at_traffic_light()
  traffic_light_state = vehicle.get_traffic_light_state().conjugate()
  return dict(
      control=control,
      location=location,
      rotation=rotation,
      velocity=velocity,
      acceleration=acceleration,
      orientation=orientation,
      angular_velocity=angular_velocity,
      speed_limit=speed_limit,
      is_at_traffic_light=is_at_traffic_light,
      traffic_light_state=traffic_light_state,
  )


def carla_xyz_to_ndarray(xyz: Any) -> np.ndarray:
  """Converts a `CARLA` measurement with attributes `x`, `y` and `z` to neural
  network friendly tensor."""
  return np.asarray(
      [xyz.x, xyz.y, xyz.z],
      dtype=np.float32,
  )


def carla_rotation_to_ndarray(
    rotation: carla.VehicleControl  # pylint: disable=no-member
) -> np.ndarray:
  """Converts a `CARLA` rotation to a neural network friendly tensor."""
  return np.asarray(
      [rotation.pitch, rotation.yaw, rotation.roll],
      dtype=np.float32,
  )


def carla_control_to_ndarray(
    control: carla.VehicleControl  # pylint: disable=no-member
) -> np.ndarray:
  """Converts a `CARLA` vehicle control to a neural network friendly tensor."""
  return np.asarray(
      [control.throttle, control.steer, control.brake],
      dtype=np.float32,
  )


def carla_measurements_to_ndarrays(
    measurements: Mapping[str, Any]) -> Mapping[str, np.ndarray]:
  """Converts the `CARLA` measurements to neural network friendly tensors."""
  control = measurements["control"]
  location = measurements["location"]
  rotation = measurements["rotation"]
  velocity = measurements["velocity"]
  acceleration = measurements["acceleration"]
  orientation = measurements["orientation"]
  angular_velocity = measurements["angular_velocity"]
  speed_limit = measurements["speed_limit"]
  is_at_traffic_light = measurements["is_at_traffic_light"]
  traffic_light_state = measurements["traffic_light_state"]
  return dict(
      control=carla_control_to_ndarray(control),
      location=carla_xyz_to_ndarray(location),
      rotation=carla_rotation_to_ndarray(rotation),
      velocity=carla_xyz_to_ndarray(velocity),
      acceleration=carla_xyz_to_ndarray(acceleration),
      orientation=carla_xyz_to_ndarray(orientation),
      angular_velocity=carla_xyz_to_ndarray(angular_velocity),
      speed_limit=np.asarray(
          speed_limit,
          dtype=np.float32,
      ),
      is_at_traffic_light=int(is_at_traffic_light),
      traffic_light_state=int(traffic_light_state),
  )


def ndarray_to_location(array: np.ndarray) -> carla.Location:  # pylint: disable=no-member
  """Converts neural network friendly tensor back to `carla.Location`."""
  return carla.Location(*list(map(float, array)))  # pylint: disable=no-member


def ndarray_to_rotation(array: np.ndarray) -> carla.Rotation:  # pylint: disable=no-member
  """Converts neural network friendly tensor back to `carla.Rotation`."""
  return carla.Rotation(*list(map(float, array)))  # pylint: disable=no-member


def ndarray_to_vector3d(array: np.ndarray) -> carla.Vector3D:  # pylint: disable=no-member
  """Converts neural network friendly tensor back to `carla.Vector3D`."""
  return carla.Vector3D(*list(map(float, array)))  # pylint: disable=no-member


def ndarray_to_control(array: np.ndarray) -> carla.VehicleControl:  # pylint: disable=no-member
  """Converts neural network friendly tensor back to `carla.VehicleControl`."""
  return carla.VehicleControl(*list(map(float, array)))  # pylint: disable=no-member


def ndarrays_to_vehicle_measurements(
    observation: Mapping[str, np.ndarray],  # pylint: disable=no-member
) -> Mapping[str, Any]:
  """Converts neural network friendly tensors back to `CARLA` objects."""
  return dict(
      control=carla.VehicleControl(*list(map(float, observation["control"]))),  # pylint: disable=no-member
      location=ndarray_to_location(observation["location"]),
      rotation=ndarray_to_rotation(observation["rotation"]),
      velocity=ndarray_to_vector3d(observation["velocity"]),
      acceleration=ndarray_to_vector3d(observation["acceleration"]),
      orientation=ndarray_to_vector3d(observation["orientation"]),
      angular_velocity=ndarray_to_vector3d(observation["angular_velocity"]),
      speed_limit=float(observation["speed_limit"]),
      is_at_traffic_light=bool(observation["is_at_traffic_light"]),
      traffic_light_state=carla.TrafficLightState.values[int(  # pylint: disable=no-member
          observation["traffic_light_state"])],
  )


def rot2mat(rotation: np.ndarray) -> np.ndarray:
  """Returns the rotation matrix (3x3) given rotation in degrees."""
  rotation_radians = ndarray_to_rotation(rotation)
  pitch = np.deg2rad(rotation_radians.pitch)
  roll = np.deg2rad(rotation_radians.roll)
  yaw = np.deg2rad(rotation_radians.yaw)
  return transforms3d.euler.euler2mat(roll, pitch, yaw).T


def world2local(*, current_location: np.ndarray, current_rotation: np.ndarray,
                world_locations: np.ndarray) -> np.ndarray:
  """Converts `world_locations` to local coordinates.

  Args:
    current_location: The ego-vehicle location, with shape `[3]`.
    current_rotation: The ego-vehicle rotation, with shape `[3]`.
    world_locations: The locations to be transformed, with shape `[..., 3]`.

  Returns:
    The local coordinates, with shape `[..., 3]`.
  """
  # Prepares interfaces.
  assert current_location.shape == (3,)
  assert current_rotation.shape == (3,)
  assert len(world_locations.shape) < 3
  world_locations = np.atleast_2d(world_locations)

  # Builds the rotation matrix.
  R = rot2mat(current_rotation)
  # Transforms world coordinates to local coordinates.
  local_locations = np.dot(a=R, b=(world_locations - current_location).T).T

  return np.squeeze(local_locations)


def local2world(*, current_location: np.ndarray, current_rotation: np.ndarray,
                local_locations: np.ndarray) -> np.ndarray:
  """Converts `local_locations` to global coordinates.

  Args:
    current_location: The ego-vehicle location, with shape `[3]`.
    current_rotation: The ego-vehicle rotation, with shape `[3]`.
    local_locations: The locations to be transformed, with shape `[..., 3]`.

  Returns:
    The global coordinates, with shape `[..., 3]`.
  """
  # Prepares interfaces.
  assert current_location.shape == (3,)
  assert current_rotation.shape == (3,)
  assert len(local_locations.shape) < 3
  local_locations = np.atleast_2d(local_locations)

  # Builds the inverse rotation matrix.
  R_inv = np.linalg.inv(rot2mat(current_rotation))
  # Transforms local coordinates to world coordinates.
  global_locations = np.dot(a=R_inv, b=local_locations.T).T + current_location

  return global_locations


def global_plan(
    world: carla.World,  # pylint: disable=no-member
    origin: carla.Location,  # pylint: disable=no-member
    destination: carla.Location,  # pylint: disable=no-member
) -> Tuple[Sequence[carla.Waypoint], Sequence[Any], float]:  # pylint: disable=no-member
  """Generates the optimal plan between two location, respecting the topology.

  Args:
    world: The `CARLA` world.
    origin: The starting location.
    destination: The final destination.

  Returns:
    waypoints: A sequence of waypoints.
    roadoptions: A sequence of commands to navigate at each waypoint.
    distances: The distance per pair of waypoints of the plan.
  """
  from agents.navigation.global_route_planner import GlobalRoutePlanner  # pylint: disable=import-error
  from agents.navigation.global_route_planner_dao import GlobalRoutePlannerDAO  # pylint: disable=import-error

  # Setup global planner.
  grp_dao = GlobalRoutePlannerDAO(wmap=world.get_map(), sampling_resolution=1)
  grp = GlobalRoutePlanner(grp_dao)
  grp.setup()
  # Generate plan.
  waypoints, roadoptions = zip(*grp.trace_route(origin, destination))
  # Accummulate pairwise distance.
  distances = [0.0]
  for i in range(1, len(waypoints)):
    loc_tm1 = waypoints[i - 1].transform.location
    loc_tm1 = np.asarray([loc_tm1.x, loc_tm1.y, loc_tm1.z])
    loc_t = waypoints[i].transform.location
    loc_t = np.asarray([loc_t.x, loc_t.y, loc_t.z])
    distances.append(np.linalg.norm(loc_tm1 - loc_t))

  return waypoints, roadoptions, distances



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines the core API for `gym.Env` and driving simulators."""

import abc
from typing import Any
from typing import Callable
from typing import Mapping
from typing import Tuple

import gym
import imageio
import numpy as np
import tqdm

from oatomobile.core.dataset import Episode
from oatomobile.core.dataset import tokens
from oatomobile.core.simulator import Action
from oatomobile.core.simulator import Observations
from oatomobile.core.simulator import Simulator
from oatomobile.util import graphics as gutil

# OpenAI Gym transition.
Transition = Tuple[Observations, float, bool, Mapping[str, Any]]


class Env(gym.Env):
  """Fundamental environment class for `oatomobile` that implements the OpenAI
  Gym interface, wrapping a driving simulator."""

  def __init__(self, sim_fn: Callable[..., Simulator], *args: Any,
               **kwargs: Any) -> None:
    """Constructs an environment, that wraps a driving simulator."""
    self._sim = sim_fn(*args, **kwargs)
    self._reset_next_step = True

  @property
  def simulator(self):
    """Returns a reference to the simulator object."""
    return self._sim

  @property
  def observation_space(self) -> gym.spaces.Dict:
    """Returns the expected observation specification returned by `step`
    method."""
    return self.simulator.observation_space

  def reset(self, *args: Any, **kwargs: Any) -> Observations:
    """Resets the state of the simulation to the initial state.

    Returns:
      The initial transition.
    """
    self._reset_next_step = False
    return self.simulator.reset(*args, **kwargs)

  def step(self, action: Action, *args: Any, **kwargs: Any) -> Transition:
    """Makes a step in the simulator, provided an action.

    Args:
      action: The hero vehicle's actions.

    Returns:
      The current timestep.
    """
    if self._reset_next_step:
      return self.reset()

    # Step the simulator.
    observation = self.simulator.step(action, *args, **kwargs)

    # Dummy return values.
    reward = 0.0
    done = False
    info = dict()

    return observation, reward, done, info

  def render(self, mode: str = "human", *args: Any, **kwargs: Any) -> Any:
    """Renders current state of the simulator."""
    return self.simulator.render(mode=mode, *args, **kwargs)

  def close(self) -> None:
    """Closes the simulator down and controls connection to CARLA server."""
    self.simulator.close()


class Metric(abc.ABC):
  """Abstract class used for metrics in `oatomobile`."""

  def __init__(self, initial_value: float, *args: Any, **kwargs: Any) -> None:
    """Initializes a stateful metric.

    Args:
      initial_value: The default/initial value of the metric.
    """
    self._initial_value = initial_value

    self.value = self._initial_value
    self.uuid = self._get_uuid(*args, **kwargs)

  def __repr__(self):
    """The universal string representation of all metrics."""
    return "{}: {}".format(self.uuid, self.value)

  def reset(self) -> None:
    """Resets `value` to initial value."""
    self.value = self._initial_value

  @abc.abstractmethod
  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the metric."""

  @abc.abstractclassmethod
  def update(self, observations: Observations, action: Action, reward: float,
             new_observations: Observations, *args: Any, **kwargs: Any) -> None:
    """Records transition and update evaluation."""


class StepsMetric(Metric):
  """Counts the number of steps in an environment."""

  def __init__(self, *args: Any, **kwargs: Any) -> None:
    """Initializes the step counter."""
    super(StepsMetric, self).__init__(initial_value=0)

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the metric."""
    return "steps"

  def update(self, observations: Observations, action: Action, reward: float,
             new_observations: Observations, *args: Any, **kwargs: Any) -> None:
    """Records transition and update evaluation."""
    self.value += 1


class ReturnsMetric(Metric):
  """Counts the cumulative undiscounted rewards in an episode."""

  def __init__(self, *args: Any, **kwargs: Any) -> None:
    """Initializes the returns counter."""
    super(ReturnsMetric, self).__init__(initial_value=0.0)

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the metric."""
    return "returns"

  def update(self, observations: Observations, action: Action, reward: float,
             new_observations: Observations, *args: Any, **kwargs: Any) -> None:
    """Records transition and update evaluation."""
    self.value += reward


class FiniteHorizonWrapper(gym.Wrapper):
  """Terminates simulation after specified number of steps."""

  def __init__(self, env: gym.Env, *, max_episode_steps: int) -> None:
    """Constructs a gym wrapper to terminate execution after
    `max_episode_steps` steps."""
    super(FiniteHorizonWrapper, self).__init__(env=env)

    # Number of steps before termination and the counter.
    self._max_episode_steps = max_episode_steps
    self._episode_step = 0
    self._pbar = tqdm.tqdm(total=self._max_episode_steps)

  def reset(self, *args: Any, **kwargs: Any) -> Observations:
    """Resets the wrapped environment and sets the counter to 0."""
    self._episode_step = 0
    self._pbar.reset()
    return self.env.reset(*args, **kwargs)

  def step(self, action: Action, *args: Any, **kwargs: Any) -> Transition:
    """Steps the wrapped environment, increments the counter and terminates if
    the maximum number of steps is reached."""
    observation, reward, done, info = self.env.step(action)
    self._episode_step += 1
    self._pbar.update(n=1)
    if self._episode_step >= self._max_episode_steps:
      done = True
    return observation, reward, done, info


class SaveToDiskWrapper(gym.Wrapper):
  """Stores observations to the disk."""

  def __init__(self, env: gym.Env, *, output_dir: str) -> None:
    """Constructs a gym wrapper to store observations to the disk."""
    super(SaveToDiskWrapper, self).__init__(env=env)

    # The parent directory to store the observations.
    self._output_dir = output_dir
    self._episode = None

  def reset(self, *args: Any, **kwargs: Any) -> Observations:
    """Resets the wrapped environment and initializes a new episode."""
    # Initializes a new episode.
    self._episode = Episode(self._output_dir, next(tokens))
    observation = self.env.reset(*args, **kwargs)
    self._episode.append(**observation)
    return observation

  def step(self, action: Action, *args: Any, **kwargs: Any) -> Transition:
    """Steps the wrapped environment, increments the counter and terminates if
    the maximum number of steps is reached."""
    observation, reward, done, info = self.env.step(action)
    self._episode.append(**observation)
    return observation, reward, done, info


class MonitorWrapper(gym.Wrapper):
  """Records a video of the episode."""

  def __init__(self,
               env: gym.Env,
               *,
               output_fname: str,
               downsample_factor: int = 1) -> None:
    """Constructs a gym wrapper to record a video of the episode."""
    super(MonitorWrapper, self).__init__(env=env)

    # The parent directory to store the video.
    self._output_fname = output_fname
    self._downsample_factor = downsample_factor
    self._recorder = imageio.get_writer(self._output_fname, mode="I")

  def reset(self, *args: Any, **kwargs: Any) -> Observations:
    """Resets the wrapped environment and initializes a new episode."""
    observation = self.env.reset(*args, **kwargs)
    self._record_frame()
    return observation

  def step(self, action: Action, *args: Any, **kwargs: Any) -> Transition:
    """Steps the wrapped environment, increments the counter and terminates if
    the maximum number of steps is reached."""
    observation, reward, done, info = self.env.step(action)
    self._record_frame()
    return observation, reward, done, info

  def _record_frame(self):
    """Appends a `frame` in the video."""
    frame = gutil.downsample(
        image=self.render(mode="rgb_array"),
        factor=self._downsample_factor,
    )
    self._recorder.append_data(frame)



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Transformations and preprocessing used across the `PyTorch` models."""

import torch
import torch.nn.functional as F

from oatomobile.core.typing import ShapeLike


def downsample_target(
    player_future: torch.Tensor,
    num_timesteps_to_keep: int,
) -> torch.Tensor:
  """Downsamples the target sequence."""
  _, T, _ = player_future.shape
  increments = T // num_timesteps_to_keep
  player_future = player_future[:, 0::increments, :]
  return player_future


def downsample_visual_features(
    visual_features: torch.Tensor,
    output_shape: ShapeLike,
) -> torch.Tensor:
  """Downsamples the visual features."""
  return F.interpolate(
      visual_features,
      size=output_shape,
      mode="bilinear",
      align_corners=True,
  )


def transpose_visual_features(visual_features: torch.Tensor) -> torch.Tensor:
  """Transposes the visual features."""
  return torch.transpose(visual_features, dim0=2, dim1=3)  # pylint: disable=no-member



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements the deep imitative model-based agent."""

from typing import Mapping

import numpy as np
import scipy.interpolate
import torch

import oatomobile
from oatomobile.baselines.base import SetPointAgent
from oatomobile.baselines.torch.dim.model import ImitativeModel


class DIMAgent(SetPointAgent):
  """The deep imitative model agent."""

  def __init__(self, environment: oatomobile.envs.CARLAEnv, *,
               model: ImitativeModel, **kwargs) -> None:
    """Constructs a deep imitation model agent.

    Args:
      environment: The navigation environment to spawn the agent.
      model: The deep imitative model.
    """
    super(DIMAgent, self).__init__(environment=environment, **kwargs)

    # Determines device, accelerator.
    self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")  # pylint: disable=no-member
    self._model = model.to(self._device)

  def __call__(self, observation: Mapping[str, np.ndarray],
               **kwargs) -> np.ndarray:
    """Returns the imitative prior."""

    # Prepares observation for the neural-network.
    observation["overhead_features"] = observation[
        "bird_view_camera_cityscapes"]
    for attr in observation:
      if not isinstance(observation[attr], np.ndarray):
        observation[attr] = np.atleast_1d(observation[attr])
      observation[attr] = observation[attr][None, ...].astype(np.float32)

    # Makes `goal` 2D.
    observation["goal"] = observation["goal"][..., :2]
    # Convert image to CHW.
    observation["lidar"] = np.transpose(observation["lidar"], (0, 3, 1, 2))
    # Processes observations for the `ImitativeModel`.
    observation = {
        key: torch.from_numpy(tensor).to(self._device)  # pylint: disable=no-member
        for (key, tensor) in observation.items()
    }
    observation = self._model.transform(observation)

    # Queries model.
    plan = self._model(num_steps=kwargs.get("num_steps", 20),
                       epsilon=kwargs.get("epsilon", 1.0),
                       lr=kwargs.get("lr", 5e-2),
                       **observation).detach().cpu().numpy()[0]  # [T, 2]

    # TODO(filangel): clean API.
    # Interpolates plan.
    player_future_length = 40
    increments = player_future_length // plan.shape[0]
    time_index = list(range(0, player_future_length, increments))  # [T]
    plan_interp = scipy.interpolate.interp1d(x=time_index, y=plan, axis=0)
    xy = plan_interp(np.arange(0, time_index[-1]))

    # Appends z dimension.
    z = np.zeros(shape=(xy.shape[0], 1))
    return np.c_[xy, z]



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Helper functions for `PyGame` and `CARLA` rendering."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import glob
import os
import tempfile
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple

import carla
import imageio
import matplotlib.pyplot as plt
import numpy as np
import pygame
import tqdm
from absl import logging
from skimage import transform


def setup(
    width: int = 400,
    height: int = 300,
    render: bool = True,
) -> Tuple[pygame.Surface, pygame.time.Clock, pygame.font.Font]:
  """Returns the `display`, `clock` and for a `PyGame` app.

  Args:
    width: The width (in pixels) of the app window.
    height: The height (in pixels) of the app window.
    render: If True it renders a window, it keeps the
      frame buffer on the memory otherwise.

  Returns:
    display: The main app window or frame buffer object.
    clock: The main app clock.
    font: The font object used for generating text.
  """
  # PyGame setup.
  pygame.init()  # pylint: disable=no-member
  pygame.display.set_caption("OATomobile")
  if render:
    logging.debug("PyGame initializes a window display")
    display = pygame.display.set_mode(  # pylint: disable=no-member
        (width, height),
        pygame.HWSURFACE | pygame.DOUBLEBUF,  # pylint: disable=no-member
    )
  else:
    logging.debug("PyGame initializes a headless display")
    display = pygame.Surface((width, height))  # pylint: disable=too-many-function-args
  clock = pygame.time.Clock()
  font = pygame.font.SysFont("dejavusansmono", 14)
  return display, clock, font


def rgb_to_binary_mask(array: np.ndarray) -> np.ndarray:
  """Flattens the `array` in the channels dimension.

  Expects black and white.
  """
  # Removes the channel dimension.
  mask = np.sum(array, axis=-1)
  # Converts array to boolean.
  return mask.astype(bool).astype(int)[..., np.newaxis]


def ndarray_to_pygame_surface(
    array: np.ndarray,
    swapaxes: bool,
) -> pygame.Surface:
  """Returns a `PyGame` surface from a `NumPy` array (image).

  Args:
    array: The `NumPy` representation of the image to be converted to `PyGame`.

  Returns:
    A `PyGame` surface.
  """
  # Make sure its in 0-255 range.
  array = 255 * (array / array.max())
  if swapaxes:
    array = array.swapaxes(0, 1)
  return pygame.surfarray.make_surface(array)


def pygame_surface_to_ndarray(surface: pygame.Surface) -> np.ndarray:
  """Returns a `NumPy` array (image) from `PyGame` surface.

  Args:
    surface: The `PyGame` surface to be converted.

  Returns:
    A `NumPy` representation of the image converted from `PyGame`.
  """
  return np.transpose(pygame.surfarray.array3d(surface), (1, 0, 2))


def mpl_figure_to_rgba(figure: plt.Figure) -> np.ndarray:
  """Returns a `NumPy` array (RGBA image) from a `matplotlib` figure.

  Args:
    figure: The `matplotlib` figure to be converted to an RGBA image.

  Returns:
    The RGBA image.
  """
  # Stores figure temporarily.
  with tempfile.NamedTemporaryFile(delete=True) as tmp:
    figure.savefig(
        "{}".format(tmp.name),
        format="png",
        bbox_inches="tight",
        pad_inches=0,
        transparent=True,
    )
    # Reads figure on memory.
    image = imageio.imread("{}".format(tmp.name))

  # Processes the image.
  image = image.astype(np.float32) / 255.0

  return image


def lidar_2darray_to_rgb(array: np.ndarray) -> np.ndarray:
  """Returns a `NumPy` array (image) from a 4 channel LIDAR point cloud.

  Args:
    array: The original LIDAR point cloud array.

  Returns:
    The `PyGame`-friendly image to be visualized.
  """
  # Get array shapes.
  W, H, C = array.shape
  assert C == 2

  # Select channel.
  img = np.c_[array, np.zeros(shape=(W, H, 1))]
  # Convert to 8-bit image.
  img = 255 * (img / img.max())
  return img


def world_to_pixel(
    location: carla.Location,  # pylint: disable=no-member
    scale: float,
    offset: Tuple[float, float],
    pixels_per_meter: int,
) -> Tuple[int, int]:
  """Converts to CARLA world coordinates to pixel coordinates.

  Args:
    world_coords: The xyz absolute (world) coordinates.
    scale: The scale factor of the coordinates.
    offset: The offset of the location.
    pixels_per_meter: The number of pixels per meter.

  Returns:
    The transformed pixel coordinates.
  """
  x = scale * pixels_per_meter * (location.x - offset[0])
  y = scale * pixels_per_meter * (location.y - offset[1])
  return (int(x - offset[0]), int(y - offset[1]))


def lateral_shift(
    transform: carla.Transform,  # pylint: disable=no-member
    shift: float,
) -> carla.Location:  # pylint: disable=no-member
  """Makes a lateral shift to the transform.

  Args:
    transform: The original transform.
    shift: The coordinate shift.

  Returns:
    The laterally shifted location.
  """
  transform.rotation.yaw += 90
  return transform.location + shift * transform.get_forward_vector()


def make_dashboard(display: pygame.Surface, font: pygame.font.Font,
                   clock: Optional[pygame.time.Clock],
                   **observations: np.ndarray) -> None:
  """Generates the dashboard used for visualizing the agent.

  Args:
    display: The `PyGame` renderable surface.
    observations: The aggregated observation object.
    font: The font object used for generating text.
    clock: The PyGame (client) clock.
  """
  # Clear dashboard.
  display.fill(COLORS["BLACK"])

  # Adaptive width.
  ada_width = 0

  if "lidar" in observations:
    # Render overhead LIDAR reading.
    ob_lidar_rgb = ndarray_to_pygame_surface(
        lidar_2darray_to_rgb(array=observations.get("lidar")),
        swapaxes=False,
    )
    display.blit(ob_lidar_rgb, (ada_width, 0))
    ada_width = ada_width + ob_lidar_rgb.get_width()

  if "left_camera_rgb" in observations:
    # Render left camera view.
    ob_left_camera_rgb_rgb = ndarray_to_pygame_surface(
        array=observations.get("left_camera_rgb"),
        swapaxes=True,
    )
    display.blit(ob_left_camera_rgb_rgb, (ada_width, 0))
    ada_width = ada_width + ob_left_camera_rgb_rgb.get_width()

  if "front_camera_rgb" in observations:
    # Render front camera view.
    ob_front_camera_rgb_rgb = ndarray_to_pygame_surface(
        array=observations.get("front_camera_rgb"),
        swapaxes=True,
    )
    display.blit(ob_front_camera_rgb_rgb, (ada_width, 0))
    ada_width = ada_width + ob_front_camera_rgb_rgb.get_width()

  if "rear_camera_rgb" in observations:
    # Render rear camera view.
    ob_rear_camera_rgb_rgb = ndarray_to_pygame_surface(
        array=observations.get("rear_camera_rgb"),
        swapaxes=True,
    )
    display.blit(ob_rear_camera_rgb_rgb, (ada_width, 0))
    ada_width = ada_width + ob_rear_camera_rgb_rgb.get_width()

  if "right_camera_rgb" in observations:
    # Render left camera view.
    ob_right_camera_rgb_rgb = ndarray_to_pygame_surface(
        array=observations.get("right_camera_rgb"),
        swapaxes=True,
    )
    display.blit(ob_right_camera_rgb_rgb, (ada_width, 0))
    ada_width = ada_width + ob_right_camera_rgb_rgb.get_width()

  for overhead_features in [
      "bird_view_camera_rgb",
      "bird_view_camera_cityscapes",
  ]:
    if overhead_features in observations:
      overhead_features_ndarray = observations.get(overhead_features)
      # Render observation.
      bev_meters = 25.0
      height, width, _ = overhead_features_ndarray.shape
      fig, ax = plt.subplots(clear=True)
      ax.imshow(
          overhead_features_ndarray,
          extent=(-bev_meters, bev_meters, bev_meters, -bev_meters),
      )
      ncol = 0
      if "goal" in observations:
        goal = observations.get("goal")
        ax.plot(
            goal[..., 1],
            -goal[..., 0],
            marker="D",
            markersize=4,
            color="#ecb01f",
            linestyle="None",
            alpha=0.25,
            label=r"$\mathcal{G}$",
        )
        ncol = ncol + 1
      if "predictions" in observations:
        predictions = observations.get("predictions")
        if predictions is not None:
          ax.plot(
              predictions[..., 1],
              -predictions[..., 0],
              marker="x",
              markersize=4,
              color="#d85218",
              alpha=0.5,
              label="agent",
          )
          ncol = ncol + 1
      if ncol > 0:
        ax.legend(
            ncol=ncol,
            loc="lower center",
            fancybox=False,
            prop={'size': 6},
        )
      ax.set(
          xlim=(-bev_meters, bev_meters),
          ylim=(bev_meters, -bev_meters),
          frame_on=False,
      )
      ax.get_xaxis().set_visible(False)
      ax.get_yaxis().set_visible(False)
      overhead_features_frame = mpl_figure_to_rgba(fig)[..., :3]
      plt.close(fig)
      overhead_features_frame = transform.resize(
          overhead_features_frame,
          output_shape=(height, width),
      )
      overhead_features_surface = ndarray_to_pygame_surface(
          overhead_features_frame,
          swapaxes=True,
      )
      display.blit(overhead_features_surface, (ada_width, 0))
      ada_width = ada_width + overhead_features_surface.get_width()

  if "bird_view_camera_cityscapes" in observations:
    # Render bird-view camera observation.
    ob_bird_view_camera_cityscapes_rgb = ndarray_to_pygame_surface(
        array=observations.get("bird_view_camera_cityscapes"),
        swapaxes=True,
    )
    display.blit(ob_bird_view_camera_cityscapes_rgb, (ada_width, 0))
    ada_width = ada_width + ob_bird_view_camera_cityscapes_rgb.get_width()

  if "control" in observations:
    # Render control bars.
    control = observations.get("control")
    background_rect = pygame.Rect(0, 180, 64, 20)
    pygame.draw.rect(display, COLORS["APPLE SPACE GREY"], background_rect, 0)
    throttle_rect = pygame.Rect(10, 190, 4, -8 * control[0])
    pygame.draw.rect(display, COLORS["GOOGLE GREEN"], throttle_rect, 0)
    steer_rect = pygame.Rect(30, 188, 8 * control[1], 4)
    pygame.draw.rect(display, COLORS["GOOGLE BLUE"], steer_rect, 0)
    break_rect = pygame.Rect(50, 190, 4, 8 * control[2])
    pygame.draw.rect(display, COLORS["GOOGLE RED"], break_rect, 0)

  # if clock is not None:
  #   text = font.render(
  #       "FPS={0:.1f}".format(clock.get_fps()),
  #       True,
  #       COLORS["WHITE"],
  #   )
  #   display.blit(text, (80, 180))

  if "velocity" in observations:
    text = font.render(
        "v={0:.1f}km/h".format(3.6 * np.linalg.norm(observations["velocity"])),
        True,
        COLORS["WHITE"],
    )
    display.blit(text, (200, 180))

  if "acceleration" in observations:
    text = font.render(
        "v={0:.1f}m/s^2".format(np.linalg.norm(observations["acceleration"])),
        True,
        COLORS["WHITE"],
    )
    display.blit(text, (300, 180))

  if "is_at_traffic_light" in observations and "traffic_light_state" in observations:
    if observations["is_at_traffic_light"] == 1 and observations[
        "traffic_light_state"] == 0:
      pygame.draw.circle(display, COLORS["GOOGLE RED"], (80, 190), 7)


def draw_settings(
    carla_map: carla.Map,  # pylint: disable=no-member
    pixels_per_meter: int = 5,
    scale: float = 1.0,
    margin: int = 150,
) -> Tuple[float, pygame.Surface]:
  """Calculates the `PyGame` surfaces settings.

  Args:
    carla_map: The `CARLA` town map.
    pixels_per_meter: The number of pixels rendered per meter.
    scale: The scaling factor of the rendered map.
    margin: The number of pixels used for margin.

  Returns:
    offset: The world offset.
    surface: The empty `PyGame` surface.
  """
  # The graph representing the CARLA map.
  waypoints = carla_map.generate_waypoints(1)

  # Calculate the width of the surface.
  max_x = max(
      waypoints,
      key=lambda x: x.transform.location.x,
  ).transform.location.x + margin
  max_y = max(
      waypoints,
      key=lambda x: x.transform.location.y,
  ).transform.location.y + margin
  min_x = min(
      waypoints,
      key=lambda x: x.transform.location.x,
  ).transform.location.x - margin
  min_y = min(
      waypoints,
      key=lambda x: x.transform.location.y,
  ).transform.location.y - margin
  world_offset = (min_x, min_y)
  width = max(max_x - min_x, max_y - min_y)
  width_in_pixels = int(pixels_per_meter * width)

  return world_offset, pygame.Surface((width_in_pixels, width_in_pixels))  # pylint: disable=too-many-function-args


def get_road_surface(
    world: carla.World,  # pylint: disable=no-member
    pixels_per_meter: int = 5,
    scale: float = 1.0,
    margin: int = 150,
) -> pygame.Surface:
  """Generates a `PyGame` surface of a CARLA town.

  Heavily inspired by the official CARLA `no_rendering_mode.py` example.

  Args:
    world: The `CARLA` world.
    pixels_per_meter: The number of pixels rendered per meter.
    scale: The scaling factor of the rendered map.
    margin: The number of pixels used for margin.

  Returns
    The topology of a CARLA town as a `PyGame` surface.
  """
  # Fetch CARLA map.
  carla_map = world.get_map()

  # Setups the `PyGame` surface and offsets.
  world_offset, surface = draw_settings(
      carla_map=carla_map,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )

  # Set background black
  surface.fill(COLORS["BLACK"])
  precision = 0.05

  # Parse OpenDrive topology.
  topology = [x[0] for x in carla_map.get_topology()]
  topology = sorted(topology, key=lambda w: w.transform.location.z)
  set_waypoints = []
  for waypoint in topology:
    waypoints = [waypoint]

    nxt = waypoint.next(precision)
    if len(nxt) > 0:
      nxt = nxt[0]
      while nxt.road_id == waypoint.road_id:
        waypoints.append(nxt)
        nxt = nxt.next(precision)
        if len(nxt) > 0:
          nxt = nxt[0]
        else:
          break
    set_waypoints.append(waypoints)

  # Draw roads.
  for waypoints in set_waypoints:
    waypoint = waypoints[0]
    road_left_side = [
        lateral_shift(w.transform, -w.lane_width * 0.5) for w in waypoints
    ]
    road_right_side = [
        lateral_shift(w.transform, w.lane_width * 0.5) for w in waypoints
    ]

    polygon = road_left_side + [x for x in reversed(road_right_side)]
    polygon = [
        world_to_pixel(
            x,
            scale=scale,
            offset=world_offset,
            pixels_per_meter=pixels_per_meter,
        ) for x in polygon
    ]

    if len(polygon) > 2:
      pygame.draw.polygon(surface, COLORS["WHITE"], polygon, 5)
      pygame.draw.polygon(surface, COLORS["WHITE"], polygon)

  return surface


def get_lane_boundaries_surface(
    world: carla.World,  # pylint: disable=no-member
    pixels_per_meter: int = 5,
    scale: float = 1.0,
    margin: int = 150,
) -> pygame.Surface:
  """Generates a `PyGame` surface of a CARLA town lane boundaries.

  Heavily inspired by the official CARLA `no_rendering_mode.py` example.

  Args:
    world: The `CARLA` world.
    pixels_per_meter: The number of pixels rendered per meter.
    scale: The scaling factor of the rendered map.
    margin: The number of pixels used for margin.

  Returns
    The lane boundaries of a CARLA town as a `PyGame` surface.
  """
  # Fetch CARLA map.
  carla_map = world.get_map()

  # Setups the `PyGame` surface and offsets.
  world_offset, surface = draw_settings(
      carla_map=carla_map,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )

  def get_lane_markings(lane_marking_type, waypoints, sign):
    margin = 0.25
    marking_1 = [
        world_to_pixel(
            lateral_shift(w.transform, sign * w.lane_width * 0.5),
            scale,
            world_offset,
            pixels_per_meter,
        ) for w in waypoints
    ]
    if lane_marking_type == carla.LaneMarkingType.Broken or (  # pylint: disable=no-member
        lane_marking_type == carla.LaneMarkingType.Solid):  # pylint: disable=no-member
      return [(lane_marking_type, marking_1)]
    else:
      marking_2 = [
          world_to_pixel(
              lateral_shift(w.transform,
                            sign * (w.lane_width * 0.5 + margin * 2)),
              scale,
              world_offset,
              pixels_per_meter,
          ) for w in waypoints
      ]
      if lane_marking_type == carla.LaneMarkingType.SolidBroken:  # pylint: disable=no-member
        return [
            (carla.LaneMarkingType.Broken, marking_1),  # pylint: disable=no-member
            (carla.LaneMarkingType.Solid, marking_2),  # pylint: disable=no-member
        ]
      elif lane_marking_type == carla.LaneMarkingType.BrokenSolid:  # pylint: disable=no-member
        return [
            (carla.LaneMarkingType.Solid, marking_1),  # pylint: disable=no-member
            (carla.LaneMarkingType.Broken, marking_2),  # pylint: disable=no-member
        ]
      elif lane_marking_type == carla.LaneMarkingType.BrokenBroken:  # pylint: disable=no-member
        return [
            (carla.LaneMarkingType.Broken, marking_1),  # pylint: disable=no-member
            (carla.LaneMarkingType.Broken, marking_2),  # pylint: disable=no-member
        ]

      elif lane_marking_type == carla.LaneMarkingType.SolidSolid:  # pylint: disable=no-member
        return [
            (carla.LaneMarkingType.Solid, marking_1),  # pylint: disable=no-member
            (carla.LaneMarkingType.Solid, marking_2),  # pylint: disable=no-member
        ]

    return [
        (carla.LaneMarkingType.NONE, []),  # pylint: disable=no-member
    ]

  def draw_solid_line(surface, color, closed, points, width):
    if len(points) >= 2:
      pygame.draw.lines(surface, color, closed, points, width)

  def draw_broken_line(surface, color, closed, points, width):
    broken_lines = [
        x for n, x in enumerate(zip(*(iter(points),) * 20)) if n % 3 == 0
    ]
    for line in broken_lines:
      pygame.draw.lines(surface, color, closed, line, width)

  def draw_lane_marking_single_side(surface, waypoints, sign):
    lane_marking = None

    marking_type = carla.LaneMarkingType.NONE  # pylint: disable=no-member
    previous_marking_type = carla.LaneMarkingType.NONE  # pylint: disable=no-member

    markings_list = []
    temp_waypoints = []
    current_lane_marking = carla.LaneMarkingType.NONE  # pylint: disable=no-member
    for sample in waypoints:
      lane_marking = sample.left_lane_marking if sign < 0 else sample.right_lane_marking

      if lane_marking is None:
        continue

      marking_type = lane_marking.type

      if current_lane_marking != marking_type:
        markings = get_lane_markings(
            previous_marking_type,
            temp_waypoints,
            sign,
        )
        current_lane_marking = marking_type

        for marking in markings:
          markings_list.append(marking)

        temp_waypoints = temp_waypoints[-1:]

      else:
        temp_waypoints.append((sample))
        previous_marking_type = marking_type

    # Add last marking.
    last_markings = get_lane_markings(
        previous_marking_type,
        temp_waypoints,
        sign,
    )
    for marking in last_markings:
      markings_list.append(marking)

    for markings in markings_list:
      if markings[0] == carla.LaneMarkingType.Solid:  # pylint: disable=no-member
        draw_solid_line(
            surface,
            COLORS["WHITE"],
            False,
            markings[1],
            2,
        )
      elif markings[0] == carla.LaneMarkingType.Broken:  # pylint: disable=no-member
        draw_broken_line(
            surface,
            COLORS["WHITE"],
            False,
            markings[1],
            2,
        )

  # Set background black
  surface.fill(COLORS["BLACK"])
  precision = 0.05

  # Parse OpenDrive topology.
  topology = [x[0] for x in carla_map.get_topology()]
  topology = sorted(topology, key=lambda w: w.transform.location.z)
  set_waypoints = []
  for waypoint in topology:
    waypoints = [waypoint]

    nxt = waypoint.next(precision)
    if len(nxt) > 0:
      nxt = nxt[0]
      while nxt.road_id == waypoint.road_id:
        waypoints.append(nxt)
        nxt = nxt.next(precision)
        if len(nxt) > 0:
          nxt = nxt[0]
        else:
          break
    set_waypoints.append(waypoints)

  # Draw roads.
  for waypoints in set_waypoints:
    waypoint = waypoints[0]
    road_left_side = [
        lateral_shift(w.transform, -w.lane_width * 0.5) for w in waypoints
    ]
    road_right_side = [
        lateral_shift(w.transform, w.lane_width * 0.5) for w in waypoints
    ]

    polygon = road_left_side + [x for x in reversed(road_right_side)]
    polygon = [
        world_to_pixel(
            x,
            scale=scale,
            offset=world_offset,
            pixels_per_meter=pixels_per_meter,
        ) for x in polygon
    ]

    # Draw Lane Markings
    if not waypoint.is_junction:
      # Left Side
      draw_lane_marking_single_side(surface, waypoints, -1)
      # Right Side
      draw_lane_marking_single_side(surface, waypoints, 1)

  return surface


def get_vehicles_surface(
    world: carla.World,  # pylint: disable=no-member
    pixels_per_meter: int = 5,
    scale: float = 1.0,
    margin: int = 150,
) -> pygame.Surface:
  """Generates a `PyGame` surface of other vehicles.

  Args:
    world: The `CARLA` world.
    pixels_per_meter: The number of pixels rendered per meter.
    scale: The scaling factor of the rendered map.
    margin: The number of pixels used for margin.

  Returns
    The vehicles rendered as a `PyGame` surface.
  """
  # Fetch CARLA map.
  carla_map = world.get_map()

  # Fetch all the vehicles.
  vehicles = [
      actor for actor in world.get_actors() if "vehicle" in actor.type_id
  ]

  # Setups the `PyGame` surface and offsets.
  world_offset, surface = draw_settings(
      carla_map=carla_map,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )

  # Set background black
  surface.fill(COLORS["BLACK"])

  # Iterate over vehicles.
  for vehicle in vehicles:
    # Draw pedestrian as a rectangular.
    corners = actor_to_corners(vehicle)
    # Convert to pixel coordinates.
    corners = [
        world_to_pixel(p, scale, world_offset, pixels_per_meter)
        for p in corners
    ]
    pygame.draw.polygon(surface, COLORS["WHITE"], corners)

  return surface


def get_pedestrians_surface(
    world: carla.World,  # pylint: disable=no-member
    pixels_per_meter: int = 5,
    scale: float = 1.0,
    margin: int = 150,
) -> pygame.Surface:
  """Generates a `PyGame` surface of pedestrians.

  Args:
    world: The `CARLA` world.
    pixels_per_meter: The number of pixels rendered per meter.
    scale: The scaling factor of the rendered map.
    margin: The number of pixels used for margin.

  Returns
    The pedestrians rendered as a `PyGame` surface.
  """
  # Fetch CARLA map.
  carla_map = world.get_map()

  # Fetch all the pedestrians.
  pedestrians = [
      actor for actor in world.get_actors()
      if "walker.pedestrian" in actor.type_id
  ]

  # Setups the `PyGame` surface and offsets.
  world_offset, surface = draw_settings(
      carla_map=carla_map,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )

  # Set background black
  surface.fill(COLORS["BLACK"])

  # Iterate over pedestrians.
  for pedestrian in pedestrians:
    # Draw pedestrian as a rectangular.
    corners = actor_to_corners(pedestrian)
    # Convert to pixel coordinates.
    corners = [
        world_to_pixel(p, scale, world_offset, pixels_per_meter)
        for p in corners
    ]
    pygame.draw.polygon(surface, COLORS["WHITE"], corners)

  return surface


def get_traffic_lights_surface(
    world: carla.World,  # pylint: disable=no-member
    pixels_per_meter: int = 5,
    scale: float = 1.0,
    margin: int = 150,
) -> Tuple[pygame.Surface, pygame.Surface, pygame.Surface]:
  """Generates three `PyGame` surfaces of traffic lights (Green, Yellow, Red).

  Args:
    world: The `CARLA` world.
    pixels_per_meter: The number of pixels rendered per meter.
    scale: The scaling factor of the rendered map.
    margin: The number of pixels used for margin.

  Returns
    The traffic lights rendered as `PyGame` surface (Green, Yellow, Red).
  """
  # Fetch CARLA map.
  carla_map = world.get_map()

  # Fetch all the pedestrians.
  traffic_lights = [
      actor for actor in world.get_actors()
      if "traffic.traffic_light" in actor.type_id
  ]

  # Setups the `PyGame` surface and offsets.
  world_offset, green_surface = draw_settings(
      carla_map=carla_map,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )
  width = green_surface.get_width()
  height = green_surface.get_height()
  yellow_surface = pygame.Surface((width, height))  # pylint: disable=too-many-function-args
  red_surface = pygame.Surface((width, height))  # pylint: disable=too-many-function-args

  # Set background black
  green_surface.fill(COLORS["BLACK"])
  yellow_surface.fill(COLORS["BLACK"])
  red_surface.fill(COLORS["BLACK"])

  # Iterate over pedestrians.
  for traffic_light in traffic_lights:
    # Identify state of the traffic light.
    if traffic_light.state.name == "Green":
      surface = green_surface
    elif traffic_light.state.name == "Yellow":
      surface = yellow_surface
    elif traffic_light.state.name == "Red":
      surface = red_surface
    else:
      continue
    # Convert to pixel coordinates.
    center = world_to_pixel(
        traffic_light.get_transform().location,
        scale,
        world_offset,
        pixels_per_meter,
    )
    pygame.draw.circle(surface, COLORS["WHITE"], center, 10)

  return green_surface, yellow_surface, red_surface


def get_hero_surface(
    world: carla.World,  # pylint: disable=no-member
    hero: carla.Vehicle,  # pylint: disable=no-member
    pixels_per_meter: int = 5,
    scale: float = 1.0,
    margin: int = 150,
) -> pygame.Surface:
  """Generates a `PyGame` surface of pedestrians.

  Args:
    world: The `CARLA` world.
    hero: The ego vehicle actor.
    pixels_per_meter: The number of pixels rendered per meter.
    scale: The scaling factor of the rendered map.
    margin: The number of pixels used for margin.

  Returns
    The ego vehicle rendered as a `PyGame` surface.
  """
  # Fetch CARLA map.
  carla_map = world.get_map()

  # Setups the `PyGame` surface and offsets.
  world_offset, surface = draw_settings(
      carla_map=carla_map,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )

  # Set background black
  surface.fill(COLORS["BLACK"])

  # Draw hero as a rectangular.
  corners = actor_to_corners(hero)
  # Convert to pixel coordinates.
  corners = [
      world_to_pixel(p, scale, world_offset, pixels_per_meter) for p in corners
  ]
  pygame.draw.polygon(surface, COLORS["WHITE"], corners)

  return surface


def actor_to_corners(actor: carla.Actor) -> Sequence[Tuple[float, float]]:  # pylint: disable=no-member
  """Draws a rectangular for the bounding box of `actor`."""
  bb = actor.bounding_box.extent
  # Draw actor as a rectangular.
  corners = [
      carla.Location(x=-bb.x, y=-bb.y),  # pylint: disable=no-member
      carla.Location(x=-bb.x, y=+bb.y),  # pylint: disable=no-member
      carla.Location(x=+bb.x, y=+bb.y),  # pylint: disable=no-member
      carla.Location(x=+bb.x, y=-bb.y),  # pylint: disable=no-member
  ]
  # Convert to global coordinates.
  actor.get_transform().transform(corners)

  return corners


def draw_game_state(
    world: carla.World,  # pylint: disable=no-member
    hero: Optional[carla.Vehicle] = None,  # pylint: disable=no-member
    pixels_per_meter: int = 5,
    scale: float = 1.0,
    margin: int = 150,
) -> pygame.Surface:
  """Generates a `PyGame` surface of the CARLA game state.

  Args:
    world: The `CARLA` world.
    hero: The ego vehicle actor.
    pixels_per_meter: The number of pixels rendered per meter.
    scale: The scaling factor of the rendered map.
    margin: The number of pixels used for margin.

  Returns
    The lane boundaries of a CARLA town as a `PyGame` surface.
  """
  # Draw individual surfaces.
  road_surface = get_road_surface(
      world=world,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )
  lane_boundaries_surface = get_lane_boundaries_surface(
      world=world,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )
  vehicles_surface = get_vehicles_surface(
      world=world,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )
  pedestrians_surface = get_pedestrians_surface(
      world=world,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )
  green_surface, yellow_surface, red_surface = get_traffic_lights_surface(
      world=world,
      pixels_per_meter=pixels_per_meter,
      scale=scale,
      margin=margin,
  )
  if hero is not None:
    hero_surface = get_hero_surface(
        world=world,
        hero=hero,
        pixels_per_meter=pixels_per_meter,
        scale=scale,
        margin=margin,
    )

  # Repaint surfaces.
  road_surface = repaint_surface(
      surface=road_surface,
      old_color=COLORS["WHITE"],
      new_color=COLORS["SILVER"],
  )
  lane_boundaries_surface = repaint_surface(
      surface=lane_boundaries_surface,
      old_color=COLORS["WHITE"],
      new_color=COLORS["GOOGLE YELLOW"],
  )
  vehicles_surface = repaint_surface(
      surface=vehicles_surface,
      old_color=COLORS["WHITE"],
      new_color=COLORS["GOOGLE BLUE"],
  )
  pedestrians_surface = repaint_surface(
      surface=pedestrians_surface,
      old_color=COLORS["WHITE"],
      new_color=COLORS["SLACK AUBERGINE"],
  )
  green_surface = repaint_surface(
      surface=green_surface,
      old_color=COLORS["WHITE"],
      new_color=COLORS["GOOGLE GREEN"],
  )
  yellow_surface = repaint_surface(
      surface=yellow_surface,
      old_color=COLORS["WHITE"],
      new_color=COLORS["GOOGLE YELLOW"],
  )
  red_surface = repaint_surface(
      surface=red_surface,
      old_color=COLORS["WHITE"],
      new_color=COLORS["GOOGLE RED"],
  )
  if hero is not None:
    hero_surface = repaint_surface(
        surface=hero_surface,
        old_color=COLORS["WHITE"],
        new_color=COLORS["SLACK BLUE"],
    )

  # Set z-ordering.
  surfaces = [
      road_surface,
      lane_boundaries_surface,
      green_surface,
      yellow_surface,
      red_surface,
      vehicles_surface,
      pedestrians_surface,
  ]

  # Append hero agent.
  if hero is not None:
    surfaces.append(hero_surface)

  # Merge surfaces.
  return merge_surfaces(
      surfaces=surfaces,
      background_color=COLORS["BLACK"],
  )


def repaint_surface(
    surface: pygame.Surface,
    old_color: pygame.Color,
    new_color: pygame.Color,
) -> pygame.Surface:
  """Replaces paints `surface` with `color`."""
  # Parses the boolean surface to a tensor.
  surf = pygame_surface_to_ndarray(surface)
  R = surf[..., 0]
  G = surf[..., 1]
  B = surf[..., 2]
  mask = (R == old_color.r) & (G == old_color.g) & (B == old_color.b)

  # Temprorary tensor.
  tmp = np.zeros(mask.shape + (3,), dtype=np.uint8)
  # Set R, G and B channels.
  tmp[mask, :3] = [new_color.r, new_color.g, new_color.b]

  return ndarray_to_pygame_surface(tmp, swapaxes=False)


def merge_surfaces(
    surfaces: Sequence[pygame.Surface],
    background_color: pygame.Color,
) -> pygame.Surface:
  """Merges all the surfaces, assuming same dimensions and z-ordering."""
  # Parse background color.
  R_B = background_color.r
  G_B = background_color.g
  B_B = background_color.b

  # Temporary tensor, used for merging.
  tmp = np.copy(pygame_surface_to_ndarray(surfaces[0]))

  # Iterate over planes.
  for surface in surfaces[1:]:
    # Parse image.
    image = pygame_surface_to_ndarray(surface)
    R = image[..., 0]
    G = image[..., 1]
    B = image[..., 2]
    # Mask of non-background pixels.
    mask = (R != R_B) + (G != G_B) + (B != B_B)
    # Overlay image.
    tmp[mask] = image[mask]

  return ndarray_to_pygame_surface(tmp, swapaxes=False)


def pngs_to_gif(dirname: str, output_filename: str) -> None:
  """Converts a list of PNGs found in `dirname` and sorted by name to a GIF."""
  pngs = sorted(glob.glob(os.path.join(dirname, "*.png")))
  with imageio.get_writer(output_filename, mode="I") as gif:
    for filename in tqdm.tqdm(pngs):
      gif.append_data(imageio.imread(filename))


def downsample(image: np.ndarray, *, factor: int) -> np.ndarray:
  """Downsamples an `image` by a `factor`."""
  return image[0::factor, 0::factor]


# Color palette, the RGB values found at https://brandpalettes.com/.
COLORS = {
    # Default palette.
    "WHITE": pygame.Color(255, 255, 255),
    "BLACK": pygame.Color(0, 0, 0),
    "RED": pygame.Color(255, 0, 0),
    "GREEN": pygame.Color(0, 255, 0),
    "BLUE": pygame.Color(0, 0, 255),
    "SILVER": pygame.Color(195, 195, 195),
    # Google palette.
    "GOOGLE BLUE": pygame.Color(66, 133, 244),
    "GOOGLE RED": pygame.Color(219, 68, 55),
    "GOOGLE YELLOW": pygame.Color(244, 160, 0),
    "GOOGLE GREEN": pygame.Color(15, 157, 88),
    # Apple palettte.
    "APPLE MIDNIGHT GREEN": pygame.Color(78, 88, 81),
    "APPLE SPACE GREY": pygame.Color(83, 81, 80),
    "APPLE ROSE GOLD": pygame.Color(250, 215, 189),
    "APPLE LIGHT PURPLE": pygame.Color(209, 205, 218),
    "APPLE LIGHT YELLOW": pygame.Color(255, 230, 129),
    "APPLE LIGHT GREEN": pygame.Color(255, 230, 129),
    "APPLE SILVER": pygame.Color(163, 170, 174),
    "APPLE BLACK": pygame.Color(31, 32, 32),
    "APPLE WHITE": pygame.Color(249, 246, 239),
    "APPLE RED": pygame.Color(165, 40, 44),
    "APPLE GOLD": pygame.Color(245, 221, 197),
    # Slack palette.
    "SLACK AUBERGINE": pygame.Color(74, 21, 75),
    "SLACK BLUE": pygame.Color(54, 197, 240),
    # Other palettes.
    "AMAZON ORANGE": pygame.Color(255, 153, 0),
    "FACEBOOK BLUE": pygame.Color(66, 103, 178),
    "AIRBNB CORAL": pygame.Color(255, 88, 93),
    "DR.PEPPER MAROON": pygame.Color(113, 31, 37),
}



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Core data structures and type definitions."""

from typing import Sequence
from typing import Union

import numpy as np


class Singleton(type):
  """Implements the singleton pattern."""

  _instances = {}

  def __call__(cls, *args, **kwargs):
    """Checks if singleton exists, creates one if not else return it."""
    if cls not in cls._instances:
      cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
    return cls._instances[cls]


Shape = Sequence[int]
ShapeLike = Union[int, Shape]
Scalar = Union[float, int]



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines the core API for a dataset."""

import abc
import os
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np

from oatomobile.util.uuid import unique_token_generator

tokens = unique_token_generator()


class Episode:
  """The abstract class for a `oatomobile` episode."""

  def __init__(
      self,
      parent_dir: str,
      token: str,
  ) -> None:
    """Constructs an episode with a unique identifier."""
    self._parent_dir = parent_dir
    self._token = token

    # The path where the samples are stored/recovered.
    self._episode_dir = os.path.join(self._parent_dir, self._token)
    os.makedirs(self._episode_dir, exist_ok=True)

    # The episode's metadata.
    self._metadata_fname = os.path.join(self._episode_dir, "metadata")

  def append(self, **observations: np.ndarray) -> None:
    """Appends `observations` to the episode."""
    # Samples a random token.
    sample_token = next(tokens)

    # Stores the `NumPy` tensors on the disk.
    np.savez_compressed(
        os.path.join(
            self._episode_dir,
            "{}.npz".format(sample_token),
        ), **observations)

    # Checks if metadata exist, else inits.
    if not os.path.exists(self._metadata_fname):
      with open(self._metadata_fname, "w") as _:
        pass
    with open(self._metadata_fname, "a") as metadata:
      metadata.write("{}\n".format(sample_token))

  def fetch(self) -> Sequence[str]:
    """Returns all the sample tokens in order."""
    with open(self._metadata_fname, "r") as metadata:
      samples = metadata.read()
    # Removes trailing newlines.
    return list(filter(None, samples.split("\n")))

  def read_sample(
      self,
      sample_token: str,
      attr: Optional[str] = None,
  ) -> Union[Mapping[str, np.ndarray], np.ndarray]:
    """Loads and parses an observation or a single attribute.

    Args:
      sample_token: The sample id of the observations.
      attr: If `None` all the attributes are loaded, else
        only the specified key is parsed and returned.

    Returns:
      Either a complete observation or a single attribute.
    """
    with np.load(
        os.path.join(
            self._episode_dir,
            "{}.npz".format(sample_token),
        ),
        allow_pickle=True,
    ) as npz_file:
      if attr is not None:
        # Returns a single attribute.
        out = npz_file[attr]
      else:
        observation = dict()
        for _attr in npz_file:
          observation[_attr] = npz_file[_attr]
        # Returns all attributes.
        out = observation

    return out


class Dataset(abc.ABC):
  """The abstract class for a `oatomobile` dataset."""

  def __init__(self, *args: Any, **kwargs: Any) -> None:
    """Constructs a dataset."""
    self.uuid = self._get_uuid(*args, **kwargs)

  @abc.abstractmethod
  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the dataset."""

  @abc.abstractproperty
  def info(self) -> Mapping[str, Any]:
    """The dataset description."""

  @abc.abstractproperty
  def url(self) -> str:
    """The URL where the dataset is hosted."""

  @abc.abstractmethod
  def download_and_prepare(self, output_dir: str, *args: Any,
                           **kwargs: Any) -> None:
    """Downloads and prepares the dataset from the host URL.

    Args:
      output_dir: The absolute path where the prepared dataset is stored.
    """

  @abc.abstractstaticmethod
  def load_datum(fname: str, *args: Any, **kwargs: Any) -> Any:
    """Loads a datum from the dataset.

    Args:
      fname: The absolute path to the datum.

    Returns:
      The parsed datum, in a Python-friendly format.
    """

  @abc.abstractstaticmethod
  def plot_datum(fname: str, output_dir: str, *args: Any,
                 **kwargs: Any) -> None:
    """Visualizes a datum from the dataset.

    Args:
      fname: The absolute path to the datum.
      output_dir: The full path to the output directory.
    """



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implementation of CARNOVEL [1], the autonomous car novel-scene benchmark to
evaluate the robustness of driving agents to a suite of tasks involving
distribution shift. CARNOVEL is based on the CARLA simulator.

#### References

[1]: Angelos Filos*, Panagiotis Tigas*, Rowan McAllister, Nicholas Rhinehart, Sergey Levine, Yarin Gal
     Can Autonomous Vehicles Identify, Recover From, and Adapt to Distribution Shifts?
"""

import functools
import glob
import json
import os
from typing import Callable
from typing import Mapping
from typing import Sequence
from typing import Text

from oatomobile.core.benchmark import Benchmark
from oatomobile.core.rl import Metric
from oatomobile.core.rl import SaveToDiskWrapper
from oatomobile.core.rl import StepsMetric
from oatomobile.envs.carla import CARLANavEnv
from oatomobile.envs.carla import CollisionsMetric
from oatomobile.envs.carla import LaneInvasionsMetric
from oatomobile.envs.carla import TerminateOnCollisionWrapper

_configs = glob.glob(
    os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "configs",
        "*.json",
    ))
_TASKS = dict()
for _config in _configs:
  _task_id = os.path.basename(_config).replace(".json", "")
  with open(_config, "r") as _dict:
    _TASKS[_task_id] = json.load(_dict)


class CORL2017(Benchmark):
  """The CORL2017 benchmark."""

  def load(self, task_id: Text, **kwargs) -> CARLANavEnv:
    """Loads a CORL2017 task.

    Args:
      task_id: The unique identifier of a task.
      max_episode_steps: The number of steps before termination, default `inf`.

    Returns:
      A task from the benchmark with `task_id`.
    """
    # TODO(filangel): figure out the correct horizon.
    env = super(CORL2017, self).load(task_id, max_episode_steps=1500, **kwargs)

    # Terminate on collision.
    env = TerminateOnCollisionWrapper(env)

    return env

  @property
  def tasks(self) -> Mapping[Text, Callable[..., CARLANavEnv]]:
    """Returns the list of tasks associated with the benchmark."""
    return {
        task_id: functools.partial(CARLANavEnv, **config)
        for (task_id, config) in _TASKS.items()
    }

  @property
  def metrics(self) -> Sequence[Metric]:
    """Returns the list of metrics associated with the benchmark."""
    return [StepsMetric(), CollisionsMetric(), LaneInvasionsMetric()]


corl2017 = CORL2017()



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines a discriminative model for the conditional imitation learner."""

from typing import Mapping
from typing import Optional
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from oatomobile.baselines.torch import transforms
from oatomobile.baselines.torch.models import MLP
from oatomobile.baselines.torch.models import MobileNetV2
from oatomobile.baselines.torch.typing import ArrayLike


class BehaviouralModel(nn.Module):
  """A `PyTorch` implementation of a behavioural cloning model."""

  def __init__(
      self,
      output_shape: Tuple[int, int] = (40, 2),
  ) -> None:
    """Constructs a simple behavioural cloning model.

    Args:
      output_shape: The shape of the base and
        data distribution (a.k.a. event_shape).
    """
    super(BehaviouralModel, self).__init__()
    self._output_shape = output_shape

    # The convolutional encoder model.
    self._encoder = MobileNetV2(num_classes=128, in_channels=2)

    # Merges the encoded features and the vector inputs.
    self._merger = MLP(
        input_size=128 + 3 + 1 + 1 + 1,
        output_sizes=[64, 64, 64],
        activation_fn=nn.ReLU,
        dropout_rate=None,
        activate_final=True,
    )

    # The decoder recurrent network used for the sequence generation.
    self._decoder = nn.GRUCell(input_size=2, hidden_size=64)

    # The output head.
    self._output = nn.Linear(
        in_features=64,
        out_features=self._output_shape[-1],
    )

  def forward(self, **context: torch.Tensor) -> torch.Tensor:
    """Returns the expert plan."""

    # Parses context variables.
    if not "visual_features" in context:
      raise ValueError("Missing `visual_features` keyword argument.")
    visual_features = context.get("visual_features")
    if not "velocity" in context:
      raise ValueError("Missing `velocity` keyword argument.")
    velocity = context.get("velocity")
    if not "is_at_traffic_light" in context:
      raise ValueError("Missing `is_at_traffic_light` keyword argument.")
    is_at_traffic_light = context.get("is_at_traffic_light")
    if not "traffic_light_state" in context:
      raise ValueError("Missing `traffic_light_state` keyword argument.")
    traffic_light_state = context.get("traffic_light_state")
    if not "mode" in context:
      raise ValueError("Missing `mode` keyword argument.")
    mode = context.get("mode")

    # Encodes the visual input.
    visual_features = self._encoder(visual_features)

    # Merges visual input logits and vector inputs.
    z = torch.cat(  # pylint: disable=no-member
        tensors=[
            visual_features,
            velocity,
            is_at_traffic_light,
            traffic_light_state,
            mode,
        ],
        dim=-1,
    )

    # The decoders initial state.
    z = self._merger(z)

    # Output container.
    y = list()

    # Initial input variable.
    x = torch.zeros(  # pylint: disable=no-member
        size=(z.shape[0], self._output_shape[-1]),
        dtype=z.dtype,
    ).to(z.device)

    # Autoregressive generation of plan.
    for _ in range(self._output_shape[0]):
      # Unrolls the GRU.
      z = self._decoder(x, z)

      # Predicts the displacement (residual).
      dx = self._output(z)
      x = dx + x

      # Updates containers.
      y.append(x)

    return torch.stack(y, dim=1)  # pylint: disable=no-member

  def transform(
      self,
      sample: Mapping[str, ArrayLike],
  ) -> Mapping[str, torch.Tensor]:
    """Prepares variables for the interface of the model.

    Args:
      sample: (keyword arguments) The raw sample variables.

    Returns:
      The processed sample.
    """

    # Preprocesses the target variables.
    if "player_future" in sample:
      sample["player_future"] = transforms.downsample_target(
          player_future=sample["player_future"],
          num_timesteps_to_keep=self._output_shape[-2],
      )

    # Renames `lidar` to `visual_features`.
    if "lidar" in sample:
      sample["visual_features"] = sample.pop("lidar")

    # Preprocesses the visual features.
    if "visual_features" in sample:
      sample["visual_features"] = transforms.transpose_visual_features(
          transforms.downsample_visual_features(
              visual_features=sample["visual_features"],
              output_shape=(100, 100),
          ))

    # Removes the "STOP" command to avoid causal confusion with traffic lights.
    if "mode" in sample:
      sample["mode"][sample["mode"] == 1.0] = 0.0

    return sample



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines a simple blind agent that follows the goals."""

import numpy as np

import oatomobile
from oatomobile.baselines.base import SetPointAgent


class BlindAgent(SetPointAgent):
  """A simple blind agent that follows the goals."""

  def __call__(self, observation: oatomobile.Observations) -> np.ndarray:
    """Returns a trajectory that connects current location and next goal."""

    return np.asarray(observation["goal"])



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines the core APIs to interface with simulators."""

import abc
from enum import Enum
from typing import Any
from typing import Iterable
from typing import Mapping
from typing import Optional
from typing import Union

import gym
import numpy as np
from absl import logging

# All agents are expected to return the same action type.
Action = Any

# Enumeration of types of sensors.
SensorTypes = Enum


class Sensor(abc.ABC):
  """A sensor consists of a fetching mechanism for observations."""

  def __init__(self, *args: Any, **kwargs: Any) -> None:
    """Constructs a sensor."""
    self.uuid = self._get_uuid(*args, **kwargs)
    self.sensor_type = self._get_sensor_type(*args, **kwargs)

  @abc.abstractmethod
  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""

  @abc.abstractmethod
  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> SensorTypes:
    """Returns the the type of the sensor."""

  @property
  @abc.abstractproperty
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Space:
    """Returns the observation spec of the sensor."""

  @abc.abstractmethod
  def get_observation(self, *args: Any, **kwargs: Any) -> Any:
    """Retrieves the observation from the sensor."""

  @abc.abstractmethod
  def close(self, *args: Any, **kwargs: Any) -> None:
    """Destroys the sensor and any connections to the server."""

  @classmethod
  @abc.abstractclassmethod
  def default(cls, *args: Any, **kwargs: Any) -> "Sensor":
    """Returns the default sensor instance."""


class Observations(dict):
  """Dictionary containing sensor observations."""

  def __init__(self, sensors: Mapping[str, Sensor], *args: Any,
               **kwargs: Any) -> None:
    """Constructors a dictionary of observations from sensors.

    Args:
      sensors: list of sensors whose observations are fetched and packaged.
    """
    data = [(uuid, sensor.get_observation(*args, **kwargs))
            for uuid, sensor in sensors.items()]
    super().__init__(data)


class SensorSuite:
  """Represents a set of sensors, with each sensor being identified through a
  unique id."""

  def __init__(
      self,
      sensors: Iterable[Sensor],
  ) -> None:
    """Constructs a sensor suite.

    Args:
      sensors: The set of all the sensors to be bundled.
    """
    self.sensors = dict()
    self._observation_space = dict()
    for sensor in sensors:
      if sensor.uuid in self.sensors:
        raise KeyError("{} is duplicated sensor uuid".format(sensor.uuid))
      self.sensors[sensor.uuid] = sensor
      self._observation_space[sensor.uuid] = sensor.observation_space

  def get(self, uuid: str) -> Sensor:
    """Returns a pointer to the sensor with uuid.

    Args:
      uuid: The universal unique identifier of the sensor.

    Returns:
      A reference to the sensor with a specific uuid.
    """
    return self.sensors.get(uuid)

  def get_observations(self, *args: Any, **kwargs: Any) -> Observations:
    """Collects data from all sensors and returns it packaged."""
    return Observations(self.sensors, *args, **kwargs)

  @property
  def observation_space(self) -> gym.spaces.Dict:
    """Returns the observation spec of the sensors."""
    return gym.spaces.Dict({
        sensor.uuid: sensor.observation_space
        for sensor in self.sensors.values()
    })

  def close(self) -> None:
    """Closes all the sensors."""
    for name, sensor in self.sensors.items():
      logging.debug("Destroys {} sensor".format(name))
      sensor.close()


class Simulator(abc.ABC):
  """Basic simulator class for `oatomobile`."""

  @property
  @abc.abstractmethod
  def sensor_suite(self) -> SensorSuite:
    """Returns a refernce to the suite of sensors."""

  @abc.abstractmethod
  def action_space(self) -> Any:
    """Returns the specification of the actions expected by the simulator."""

  @property
  def observation_space(self) -> gym.spaces.Dict:
    """Returns the specification of the observations returned by the
    simulator."""
    return self.sensor_suite.observation_space

  @abc.abstractmethod
  def seed(self, seed: int) -> None:
    """Fixes the random number generator state."""

  @abc.abstractmethod
  def reset(self, *args: Any, **kwargs: Any) -> Observations:
    """Resets the state of the simulation to the initial state."""

  @abc.abstractmethod
  def step(self, action: Action, *args: Any, **kwargs: Any) -> Observations:
    """Makes a step in the simulator, provided an action."""

  @abc.abstractmethod
  def render(self, mode: str = "rgb_array", *args: Any, **kwargs: Any) -> Any:
    """Renders current state of the simulator."""

  @abc.abstractmethod
  def close(self) -> None:
    """Closes the simulator down and controls connection to the server."""



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines the core API for benchmarks within `oatomobile`."""

import abc
import functools
import os
from typing import Any
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Union

import pandas as pd
import tqdm
from absl import logging

from oatomobile.core.agent import Agent
from oatomobile.core.loop import EnvironmentLoop
from oatomobile.core.rl import Env
from oatomobile.core.rl import FiniteHorizonWrapper
from oatomobile.core.rl import Metric
from oatomobile.core.rl import MonitorWrapper
from oatomobile.core.typing import Scalar


class Benchmark(abc.ABC):
  """An abstract class for a benchmark in `oatomobile`."""

  @abc.abstractproperty
  def metrics(self) -> Sequence[Metric]:
    """Returns the list of metrics associated with the benchmark."""

  @abc.abstractproperty
  def tasks(self) -> Mapping[str, Callable[..., Env]]:
    """Returns the list of tasks associated with the benchmark."""

  def load(self,
           task_id: str,
           max_episode_steps: Optional[int] = None,
           *args: Any,
           **kwargs: Any) -> Env:
    """Loads a task.

    Args:
      task_id: The unique identifier of a task.
      max_episode_steps: The number of steps before termination, default `inf`.

    Returns:
      A task from the benchmark with `task_id`.
    """
    if task_id not in self.tasks:
      raise ValueError("Unrecognised task with id {}".format(task_id))

    # Initializes environment.
    env = self.tasks[task_id](*args, **kwargs)

    if max_episode_steps is not None:
      env = FiniteHorizonWrapper(env, max_episode_steps=max_episode_steps)

    return env

  def evaluate(self,
               agent_fn: Callable[..., Agent],
               log_dir: str,
               render: bool = False,
               monitor: bool = False,
               subtasks_id: Optional[str] = None,
               *args: Any,
               **kwargs: Any) -> None:
    """Runs a full evaluation of an agent on the benchmark.

    Args:
      agent_fn: The agent's construction function that receives each task.
      log_dir: The full path to the directory where all the logs are kept.
      render: If True, it renders the display.
      monitor: If True, it stores the videos on the screen.
      subtasks_id: The subset of tasks to run, matches regex.
    """
    # Makes sure the output directory exists.
    os.makedirs(log_dir, exist_ok=True)

    # Keep only the tasks that have `subtasks`.
    tasks = self.tasks if subtasks_id is None else [
        task for task in self.tasks if subtasks_id in task
    ]

    # Evaluate on tasks, sequentially -- could be run on parallel too.
    for task_id in tqdm.tqdm(tasks):
      logging.debug("Start evaluation on task {}".format(task_id))
      task_dir = os.path.join(log_dir, task_id)
      os.makedirs(task_dir, exist_ok=True)

      # Load environment.
      env = self.load(task_id)
      if monitor:
        video_fname = os.path.join(task_dir, "video.gif")
        env = MonitorWrapper(env, output_fname=video_fname)

      # Run episode and record metrics.
      results = EnvironmentLoop(
          agent_fn=functools.partial(agent_fn, *args, **kwargs),
          environment=env,
          metrics=self.metrics,
          render_mode="human" if render else "none",
      ).run()

      # Dumps results in a CSV file.
      results = {uuid: [value] for (uuid, value) in results.items()}
      pd.DataFrame(results).to_csv(
          os.path.join(task_dir, "metrics.csv"),
          header=True,
          index=False,
      )



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Helper functions for logging, saving and loading models in `PyTorch`."""

import os
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
import torchvision
from torch.utils.tensorboard import SummaryWriter

from oatomobile.baselines.torch.typing import ArrayLike

COLORS = [
    "#0071bc",
    "#d85218",
    "#ecb01f",
    "#7d2e8d",
    "#76ab2f",
    "#4cbded",
    "#a1132e",
]

# Determines device, accelerator.
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")  # pylint: disable=no-member


class Checkpointer:
  """A simple `PyTorch` model load/save wrapper."""

  def __init__(
      self,
      model: nn.Module,
      ckpt_dir: str,
  ) -> None:
    """Constructs a simple load/save checkpointer."""
    self._model = model
    self._ckpt_dir = ckpt_dir
    os.makedirs(self._ckpt_dir, exist_ok=True)

  def save(
      self,
      epoch: int,
  ) -> str:
    """Saves the model to the `ckpt_dir/epoch/model.pt` file."""
    ckpt_path = os.path.join(self._ckpt_dir, "model-{}.pt".format(epoch))
    torch.save(self._model.state_dict(), ckpt_path)
    return ckpt_path

  def load(
      self,
      epoch: int,
  ) -> nn.Module:
    """Loads the model from the `ckpt_dir/epoch/model.pt` file."""
    ckpt_path = os.path.join(self._ckpt_dir, "model-{}.pt".format(epoch))
    self._model.load_state_dict(torch.load(ckpt_path, map_location=device))
    return self._model


class TensorBoardWriter:
  """A simple `Pytorch`-friendly `TensorBoard` wrapper."""

  def __init__(
      self,
      log_dir: str,
  ) -> None:
    """Constructs a simgple `TensorBoard` wrapper."""
    # Makes sure output directories exist.
    log_dir_train = os.path.join(log_dir, "train")
    log_dir_val = os.path.join(log_dir, "val")
    os.makedirs(log_dir_train, exist_ok=True)
    os.makedirs(log_dir_val, exist_ok=True)

    # Initialises the `TensorBoard` writters.
    self._summary_writter_train = SummaryWriter(log_dir=log_dir_train)
    self._summary_writter_val = SummaryWriter(log_dir=log_dir_val)

  def log(
      self,
      split: str,
      loss: float,
      overhead_features: ArrayLike,
      predictions: ArrayLike,
      ground_truth: ArrayLike,
      global_step: int,
  ) -> None:
    """Logs the scalar loss and visualizes predictions for qualitative
    inspection."""

    if split == "train":
      summary_writter = self._summary_writter_train
    elif split == "val":
      summary_writter = self._summary_writter_val
    else:
      raise ValueError("Unrecognised split={} was passed".format(split))

    # Logs the training loss.
    summary_writter.add_scalar(
        tag="loss",
        scalar_value=loss,
        global_step=global_step,
    )

    # Visualizes the predictions.
    overhead_features = np.transpose(overhead_features, (0, 2, 3, 1))  # to NHWC
    raw = list()
    for t, (
        o_t,
        p_t,
        g_t,
    ) in enumerate(zip(
        overhead_features,
        predictions,
        ground_truth,
    )):
      fig, ax = plt.subplots(figsize=(3.0, 3.0))
      bev_meters = 25.0
      # Overhead features.
      ax.imshow(
          o_t.squeeze()[..., 0],
          extent=(-bev_meters, bev_meters, bev_meters, -bev_meters),
          cmap="gray",
      )
      # Ground truth.
      ax.plot(
          g_t[..., 1],
          -g_t[..., 0],
          marker="o",
          markersize=4,
          color=COLORS[1],
          alpha=1.0,
          label="ground truth",
      )
      # Model prediction.
      ax.plot(
          p_t[..., 1],
          -p_t[..., 0],
          marker="o",
          markersize=4,
          color=COLORS[0],
          alpha=0.75,
          label="predictions",
      )
      ax.legend()
      ax.set(frame_on=False)
      ax.get_xaxis().set_visible(False)
      ax.get_yaxis().set_visible(False)
      # Convert `matplotlib` canvas to `NumPy` RGB.
      fig.canvas.draw()
      w, h = fig.canvas.get_width_height()
      buf = np.fromstring(fig.canvas.tostring_argb(), dtype=np.uint8)
      buf.shape = (w, h, 4)
      buf = np.roll(buf, 3, axis=2)
      raw.append(buf)
      plt.close(fig)
    raw = np.reshape(np.asarray(raw), (-1, w, h, 4))
    summary_writter.add_images(
        tag="examples",
        img_tensor=raw[..., :3],
        global_step=global_step,
        dataformats="NHWC",
    )
